import pytest
import fast_carpenter.backends as backends


def test_get_backend():
    alphatwirl_back = backends.get_backend("alphatwirl:multiprocessing")
    assert hasattr(alphatwirl_back, "execute")

    with pytest.raises(ValueError) as e:
        backends.get_backend("doesn't exist")
    assert "Unknown backend" in str(e)



import pytest
import six
import fast_carpenter.selection.stage as stage
from fast_carpenter.masked_tree import MaskedUprootTree
import fast_carpenter.selection.filters as filters
from ..conftest import FakeBEEvent


def test__create_weights_none():
    weights = stage._create_weights("test__create_weights_none", None)
    assert isinstance(weights, dict)
    assert len(weights) == 0


def test__create_weights_single_string():
    weights = "dummy_weight"
    weights = stage._create_weights("test__create_weights_none", weights)
    assert isinstance(weights, dict)
    assert len(weights) == 1
    assert list(weights.keys()) == ["dummy_weight"]
    assert all([k == v for k, v in weights.items()])


def test__create_weights_list():
    weights = ["hello", "world"]
    weights = stage._create_weights("test__create_weights_none", weights)
    assert isinstance(weights, dict)
    assert len(weights) == 2
    assert list(sorted(weights.keys())) == sorted(["hello", "world"])
    assert all([k == v for k, v in weights.items()])


def test__create_weights_dict():
    weights = {"HEY": "hello", "EARTH": "world"}
    weights = stage._create_weights("test__create_weights_none", weights)
    assert isinstance(weights, dict)
    assert len(weights) == 2
    assert list(sorted(weights.keys())) == sorted(["HEY", "EARTH"])
    assert list(sorted(weights.values())) == sorted(["hello", "world"])


@pytest.fixture
def cutflow_1(tmpdir):
    return stage.CutFlow("cutflow_1", str(tmpdir), selection="NMuon > 1", weights="NElectron")


def test_cutflow_1(cutflow_1):
    assert cutflow_1._weights == {"NElectron": "NElectron"}
    assert isinstance(cutflow_1.selection, filters.OuterCounterIncrementer)
    assert isinstance(cutflow_1.selection._wrapped_selection, filters.SingleCut)
    assert isinstance(cutflow_1.selection.selection, six.string_types)


def test_cutflow_1_executes_mc(cutflow_1, infile, full_event_range, tmpdir):
    chunk = FakeBEEvent(MaskedUprootTree(infile, event_ranger=full_event_range), "mc")
    cutflow_1.event(chunk)

    assert len(chunk.tree) == 289

    collector = cutflow_1.collector()
    assert collector.filename == str(tmpdir / "cuts_cutflow_1-NElectron.csv")

    dataset_readers_list = (("test_mc", (cutflow_1, )), )
    output = collector._prepare_output(dataset_readers_list)
    assert len(output) == 1
    assert all(output[("passed_only_cut", "unweighted")] == [289])
    assert all(output[("passed_incl", "unweighted")] == [289])
    assert all(output[("totals_incl", "unweighted")] == [4580])


def test_cutflow_1_executes_data(cutflow_1, infile, full_event_range, tmpdir):
    chunk = FakeBEEvent(MaskedUprootTree(infile, event_ranger=full_event_range), "data")
    cutflow_1.event(chunk)

    assert len(chunk.tree) == 289

    collector = cutflow_1.collector()
    assert collector.filename == str(tmpdir / "cuts_cutflow_1-NElectron.csv")


@pytest.fixture
def select_2(tmpdir):
    select = {"All": ["NMuon > 1",
                      {"Any": ["NElectron > 1", "NJet > 1"]},
                      {"reduce": 1, "formula": "Muon_Px > 0.3"}]}
    return select


def one_stage_many_calls(selection, tmpdir, chunk_data, chunk_mc):
    cutflow_a = stage.CutFlow("cutflow_a", str(tmpdir), selection=selection, weights="EventWeight")
    cutflow_b = stage.CutFlow("cutflow_b", str(tmpdir), selection=selection, weights="EventWeight")

    cutflow_a.event(chunk_mc)
    chunk_mc.tree.reset_mask()
    cutflow_a.event(chunk_mc)

    cutflow_b.event(chunk_data)
    chunk_data.tree.reset_mask()
    cutflow_b.event(chunk_data)

    collector = cutflow_a.collector()
    dataset_readers_list = (("test_mc", (cutflow_a,)), ("test_data", (cutflow_b,)),)
    return collector, dataset_readers_list


def many_stages_one_call(selection, tmpdir, chunk_data, chunk_mc):
    cutflows = [stage.CutFlow("cutflow_%d" % i, str(tmpdir), selection=selection, weights="EventWeight")
                for i in range(4)]
    cutflows[0].event(chunk_mc)
    chunk_mc.tree.reset_mask()
    cutflows[1].event(chunk_mc)
    cutflows[2].event(chunk_data)
    chunk_data.tree.reset_mask()
    cutflows[3].event(chunk_data)

    collector = cutflows[0].collector()
    dataset_readers_list = (("test_mc", cutflows[:2]), ("test_data", cutflows[2:]),)
    return collector, dataset_readers_list


@pytest.mark.parametrize("multi_chunk_func", [one_stage_many_calls, many_stages_one_call])
def test_cutflow_2_collect(select_2, tmpdir, infile, full_event_range, multi_chunk_func):
    chunk_data = FakeBEEvent(MaskedUprootTree(infile, event_ranger=full_event_range), "data")
    chunk_mc = FakeBEEvent(MaskedUprootTree(infile, event_ranger=full_event_range), "mc")

    collector, dataset_readers_list = multi_chunk_func(select_2, tmpdir, chunk_data, chunk_mc)
    output = collector._prepare_output(dataset_readers_list)

    assert len(output) == 12
    data = output.xs("test_data", level="dataset", axis="rows")
    data_weighted = data.xs("EventWeight", level=1, axis="columns")
    data_unweighted = data.xs("unweighted", level=1, axis="columns")
    assert all(data_weighted == data_unweighted)
    mc = output.xs("test_mc", level="dataset", axis="rows")
    mc_unweighted = mc.xs("unweighted", level=1, axis="columns")
    assert all(mc_unweighted == data_unweighted)
    assert output.loc[("test_data", 0, "All"), ("totals_incl", "unweighted")] == 4580 * 2
    assert output.loc[("test_data", 0, "All"), ("totals_incl", "EventWeight")] == 4580 * 2
    assert output.loc[("test_mc", 0, "All"), ("totals_incl", "unweighted")] == 4580 * 2
    assert output.loc[("test_data", 1, "NMuon > 1"), ("passed_only_cut", "unweighted")] == 289 * 2
    assert output.loc[("test_mc", 1, "NMuon > 1"), ("passed_only_cut", "unweighted")] == 289 * 2

    coll_out = collector.collect(dataset_readers_list, writeFiles=False)

    assert len(coll_out) == 12
    data = coll_out.xs("test_data", level="dataset", axis="rows")
    data_weighted = data.xs("EventWeight", level=1, axis="columns")
    data_unweighted = data.xs("unweighted", level=1, axis="columns")
    assert all(data_weighted == data_unweighted)
    mc = coll_out.xs("test_mc", level="dataset", axis="rows")
    mc_unweighted = mc.xs("unweighted", level=1, axis="columns")
    assert all(mc_unweighted == data_unweighted)
    assert coll_out.loc[("test_data", 0, "All"), ("totals_incl", "unweighted")] == 4580 * 2
    assert coll_out.loc[("test_data", 0, "All"), ("totals_incl", "EventWeight")] == 4580 * 2
    assert coll_out.loc[("test_mc", 0, "All"), ("totals_incl", "unweighted")] == 4580 * 2
    assert coll_out.loc[("test_data", 1, "NMuon > 1"), ("passed_only_cut", "unweighted")] == 289 * 2
    assert coll_out.loc[("test_mc", 1, "NMuon > 1"), ("passed_only_cut", "unweighted")] == 289 * 2


def test_sequential_stages(cutflow_1, select_2, infile, full_event_range, tmpdir):
    cutflow_2 = stage.CutFlow("cutflow_2", str(tmpdir), selection=select_2, weights="EventWeight")
    chunk = FakeBEEvent(MaskedUprootTree(infile, event_ranger=full_event_range), "data")
    cutflow_1.event(chunk)
    cutflow_2.event(chunk)

    assert len(chunk.tree) == 2
    jet_py = chunk.tree.array("Jet_Py")
    assert pytest.approx(jet_py.flatten()) == [49.641838, 45.008915, -78.01798, 60.730812]



import pytest
import six
import numpy as np
import uproot3
import fast_carpenter.selection.filters as filters


@pytest.fixture
def filename():
    return "tests/data/CMS_HEP_tutorial_ww.root"


@pytest.fixture
def config_1():
    return "NMuon > 1"


def test_build_selection_1(config_1):
    selection = filters.build_selection("test_build_selection_1", config_1)
    assert isinstance(selection, filters.OuterCounterIncrementer)
    assert isinstance(selection._wrapped_selection, filters.SingleCut)
    assert isinstance(selection.selection, six.string_types)


def test_selection_1(config_1, filename):
    selection = filters.build_selection("test_selection_1", config_1)
    infile = uproot3.open(filename)["events"]
    mask = selection(infile, is_mc=False)
    assert np.count_nonzero(mask) == 289

    columns = selection.columns
    values = selection.values
    index = selection.index_values
    assert len(values) == 1
    assert len(index) == 1
    assert index[0][0] == "0"
    assert index[0][1] == 0
    assert index[0][2] == "NMuon > 1"
    assert values[0][columns[0].index("passed_incl")] == 289
    assert values[0][columns[0].index("passed_only_cut")] == 289
    assert values[0][columns[0].index("totals_incl")] == 4580

    df = selection.to_dataframe()
    assert len(df) == 1
    assert all(df.index.get_level_values("depth") == [0])
    assert all(df.index.get_level_values("cut") == ["NMuon > 1"])
    row = ("0", 0, "NMuon > 1")
    assert df.loc[row, ("passed_incl", "unweighted")] == 289
    assert df.loc[row, ("passed_only_cut", "unweighted")] == 289
    assert df.loc[row, ("totals_incl", "unweighted")] == 4580


@pytest.fixture
def config_2():
    return {"Any": ["NMuon > 1", "NElectron > 1", "NJet > 1"]}


def test_selection_2_weights(config_2, filename):
    selection = filters.build_selection("test_selection_1",
                                        config_2, weights=["EventWeight"])
    infile = uproot3.open(filename)["events"]
    mask = selection(infile, is_mc=True)
    assert np.count_nonzero(mask) == 1486

    columns = selection.columns
    assert len(columns) == 2
    assert len(columns[0]) == 6

    values = selection.values
    index = selection.index_values
    assert len(values) == 4
    assert len(index) == 4
    assert index[0] == ("0", 0, "Any")
    assert values[0][columns[0].index("passed_incl")] == 1486
    assert values[0][columns[0].index("passed_incl") + 1] == np.sum(infile.array("EventWeight")[mask])


def test_selection_2_weights_data(config_2, filename):
    selection = filters.build_selection("test_selection_1",
                                        config_2, weights=["EventWeight"])
    infile = uproot3.open(filename)["events"]
    mask = selection(infile, is_mc=False)
    assert np.count_nonzero(mask) == 1486

    columns = selection.columns
    assert len(columns) == 2
    assert len(columns[0]) == 6

    index = selection.index_values
    values = selection.values

    assert len(values) == 4
    assert len(index) == 4
    assert index[0] == ("0", 0, "Any")
    assert index[1] == ("0,0", 1, "NMuon > 1")
    assert values[0][columns[0].index("passed_incl")] == 1486
    assert values[0][columns[0].index("passed_incl") + 1] == 1486


@pytest.fixture
def config_3():
    return {"All": ["NMuon > 1", {"Any": ["NElectron > 1", "NJet > 1"]}]}


def test_selection_3(config_3, filename):
    selection = filters.build_selection("test_selection_3", config_3)
    infile = uproot3.open(filename)["events"]
    mask = selection(infile, is_mc=True)
    assert np.count_nonzero(mask) == 8

    index = selection.index_values
    values = selection.values
    columns = selection.columns
    assert len(values) == 5
    assert index[0] == ("0", 0, "All")
    assert index[1] == ("0,0", 1, "NMuon > 1")
    assert values[0][columns[0].index("passed_incl")] == 8
    assert values[1][columns[0].index("passed_incl")] == 289


@pytest.fixture
def config_jagged_index():
    return dict(reduce=1, formula="Muon_Px > 0.3")


def test_selection_jagged_index(config_jagged_index, filename):
    selection = filters.build_selection("test_selection_jagged", config_jagged_index)
    infile = uproot3.open(filename)["events"]
    mask = selection(infile, is_mc=False)
    # Compare to: events->Draw("", "Muon_Px[1] > 0.300")
    assert len(mask) == len(infile)
    assert np.count_nonzero(mask) == 144


@pytest.fixture
def config_jagged_count_nonzero():
    return dict(reduce="any", formula="Muon_Px > 0.3")


def test_selection_jagged_count_nonzero(config_jagged_count_nonzero, filename):
    selection = filters.build_selection("test_selection_jagged", config_jagged_count_nonzero)
    infile = uproot3.open(filename)["events"]
    mask = selection(infile, is_mc=False)
    # Compare to: events->Draw("", "Sum$(Muon_Px > 0.300) > 0")
    assert np.count_nonzero(mask) == 2225


def fake_evaluate(variables, expression):
    import numexpr
    return numexpr.evaluate(expression, variables)


class FakeTree(dict):
    def __init__(self):
        super(FakeTree, self).__init__(NMuon=np.linspace(0, 5., 101),
                                       NElectron=np.linspace(0, 10, 101),
                                       NJet=np.linspace(2, -18, 101),
                                       )

    def __len__(self):
        return 101


def test_event_removal_1(config_1, monkeypatch):
    variables = FakeTree()
    monkeypatch.setattr(filters, 'evaluate', fake_evaluate)
    selection = filters.build_selection("test_event_removal", config_1)
    nmuon = variables["NMuon"]

    mask = selection(variables, is_mc=False)
    assert len(mask) == 101
    assert mask.dtype.kind == "b"
    assert len(nmuon[mask]) == 80
    assert nmuon[mask].max() == 5
    assert nmuon[mask].min() == 1.05


def test_event_removal_2(config_2, monkeypatch):
    variables = FakeTree()
    monkeypatch.setattr(filters, 'evaluate', fake_evaluate)
    selection = filters.build_selection("test_event_removal", config_2)

    mask = selection(variables, is_mc=False)

    assert len(mask) == 101
    assert mask.dtype.kind == "b"
    assert np.count_nonzero(mask) == 95
    assert all(mask == [True] * 5 + [False] * 6 + [True] * 90)
    assert variables["NMuon"][mask].min() == 0
    assert variables["NElectron"][mask].min() == 0
    assert variables["NJet"][mask].min() == -18


def test_event_removal_3(config_3, monkeypatch):
    variables = FakeTree()
    monkeypatch.setattr(filters, 'evaluate', fake_evaluate)
    selection = filters.build_selection("test_event_removal", config_3)

    mask = selection(variables, is_mc=False)

    assert len(mask) == 101
    assert mask.dtype.kind == "b"
    assert np.count_nonzero(mask) == 80
    assert all(mask == [False] * 21 + [True] * 80)
    assert variables["NMuon"][mask].min() == 1.05
    assert variables["NElectron"][mask].min() == 2.1
    assert variables["NJet"][mask].max() == -2.2
    assert variables["NJet"][mask].min() == -18



# -*- coding: utf-8 -*-
#
# fast-carpenter documentation build configuration file, created by
# sphinx-quickstart on Thu Jul 13 09:48:02 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import inspect
import fast_carpenter


# Get the project root dir, which is the parent dir of this
cwd = os.getcwd()
project_root = os.path.dirname(cwd)
# Insert the project root dir as the first element in the PYTHONPATH.
# This lets us ensure that the source package is imported, and that its
# version is used.
sys.path.insert(0, project_root)


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinxcontrib.apidoc',
    'sphinxcontrib.programoutput',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.linkcode',
    'sphinx.ext.napoleon',
    # 'sphinx.ext.githubpages',
    'm2r',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = ['.rst', '.md']
#source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'fast-carpenter'
copyright = u'2019, Ben Krikler'
author = u'Ben Krikler'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
# version = u'0.7'
# The full version, including alpha/beta/rc tags.
release = fast_carpenter.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# List of modules to be mocked up. This is useful when some external
# dependencies are not met at build time and break the building process.
autodoc_mock_imports = [
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

html_theme = 'sphinx_rtd_theme'

# Increase width of main content by overriding corresponding parameter in rtd theme's css
#   Solution from http://blog.deimos.fr/2014/10/02/sphinxdoc-and-readthedocs-theme-tricks-2/
#            and  http://stackoverflow.com/questions/23211695/modifying-content-width-of-the-sphinx-theme-read-the-docs
import sphinx_rtd_theme
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

#html_theme = 'alabaster'
#
# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
#html_theme_options = {
#    'page_width': '1200px',
#    'show_related': 'true',
#    'fixed_sidebar': 'true',
#    'font_family': 'Georgia'
#}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'fast-carpenter-doc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'fast-carpenter.tex', u'fast-carpenter Documentation',
     [author], 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'fast-carpenter', u'fast-carpenter Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'fast-carpenter', u'fast-carpenter Documentation',
     author, 'fast-carpenter', 'One line description of project.',
     'Miscellaneous'),
]


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'python': ('https://docs.python.org/3', None),
                       'matplotlib': ('http://matplotlib.sourceforge.net', None),
                       'pandas': ('http://pandas.pydata.org/pandas-docs/stable', None),
                      }


# Configure the sphinxcontrib.apidoc extension
apidoc_module_dir = '../fast_carpenter'
apidoc_toc_file = False
apidoc_module_first = True
apidoc_separate_modules = True
pidoc_extra_args = ["-e"]


# linkcode control, inspired by numpy's approach:
#   - https://github.com/numpy/numpy/blob/master/doc/source/conf.py#L286
def linkcode_resolve(domain, info):
    if domain != 'py':
        return None

    modname = info['module']
    fullname = info['fullname']

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split('.'):
        try:
            obj = getattr(obj, part)
        except Exception:
            return None

    # strip decorators, which would resolve to the source of the decorator
    # possibly an upstream bug in getsourcefile
    try:
        unwrap = inspect.unwrap
    except AttributeError:
        pass
    else:
        obj = unwrap(obj)

    try:
        fn = inspect.getsourcefile(obj)
    except Exception:
        fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except Exception:
        lineno = None

    if lineno:
        linespec = "#L%d-L%d" % (lineno, lineno + len(source) - 1)
    else:
        linespec = ""

    fn = os.path.relpath(fn, start=os.path.dirname(fast_carpenter.__file__))
    version = os.environ.get("READTHEDOCS_VERSION", "latest")
    if version == "latest":
        tgt_version = "master"
    elif version == "stable":
        tgt_version = "v" + fast_carpenter.__version__
    else:
        tgt_version = version
    url_root = "https://github.com/FAST-HEP/fast-carpenter/tree/"
    return url_root + "%s/fast_carpenter/%s%s" % (tgt_version, fn, linespec)


# Control the napoleon extension for nicer docstrings
# napoleon_include_init_with_doc = True



# -- Modify sphinx app -------------------------------------------
#from docutils.parsers.rst import Directive, directives
#class AutodocWithoutDocstring(Directive):
#
#    from sphinx.ext.autodoc import cut_lines
#    first_line = cut_lines(1)
#    print(options)
#    if "first-line" in options:
#        del options["first-line"]
#        first_line(app, what, name, obj, options, lines)

def setup(app):
    # Override default css to get a larger width for local build
    #app.add_javascript("custom.js")
    app.add_stylesheet('custom.css')
    # app.connect('autodoc-without-docstring', AutodocWithoutDocstring)




import pytest
import uproot3
from collections import namedtuple


@pytest.fixture
def infile():
    filename = "tests/data/CMS_HEP_tutorial_ww.root"
    return uproot3.open(filename)["events"]


FakeEventRange = namedtuple("FakeEventRange", "start_entry stop_entry entries_in_block")


@pytest.fixture
def event_range():
    return FakeEventRange(100, 200, 100)


@pytest.fixture
def full_event_range():
    return FakeEventRange(0, 4580, 0)


@pytest.fixture
def wrapped_tree(infile, event_range):
    import fast_carpenter.tree_wrapper as tree_w
    tree = tree_w.WrappedTree(infile, event_range)
    return tree


@pytest.fixture
def full_wrapped_tree(infile, full_event_range):
    import fast_carpenter.tree_wrapper as tree_w
    tree = tree_w.WrappedTree(infile, full_event_range)
    return tree


class Namespace():
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class FakeBEEvent(object):
    def __init__(self, tree, eventtype):
        self.tree = tree
        self.config = Namespace(dataset=Namespace(eventtype=eventtype))



import pytest
import numpy as np


def test_add_retrieve(wrapped_tree):
    muon_px = wrapped_tree.array("Muon_Px")
    assert len(muon_px) == 100

    muon_py, muon_pz = wrapped_tree.arrays(["Muon_Py", "Muon_Pz"], outputtype=tuple)
    muon_momentum = np.hypot(muon_py, muon_pz)
    wrapped_tree.new_variable("Muon_momentum", muon_momentum)
    retrieve_momentum = wrapped_tree.array("Muon_momentum")
    assert (retrieve_momentum == muon_momentum).flatten().all()


def test_overwrite(wrapped_tree):
    muon_px = wrapped_tree.array("Muon_Px")
    with pytest.raises(ValueError) as err:
        wrapped_tree.new_variable("Muon_Px", muon_px / muon_px)
    assert "Muon_Px" in str(err)
    assert len(wrapped_tree.keys(filtername=lambda x: x.decode() == "Muon_Px")) == 1



import pandas as pd
import copy
import numpy as np
import pytest
import fast_carpenter.summary.binned_dataframe as bdf
from . import dummy_binning_descriptions as binning
from ..conftest import FakeBEEvent


@pytest.fixture
def config_1():
    return dict(binning=[binning.bins_met_px,
                         binning.bins_py],
                weights=binning.weight_list)


@pytest.fixture
def config_2():
    return dict(binning=[binning.bins_py,
                         binning.bins_met_px,
                         binning.bins_nmuon],
                weights=binning.weight_dict)


@pytest.fixture
def config_3():
    return dict(binning=[binning.bins_electron_pT],
                weights=binning.weight_dict)


@pytest.fixture
def binned_df_1(config_1):
    return make_binned_df_1(config_1)


def make_binned_df_1(config_1):
    return bdf.BinnedDataframe("binned_df_1", out_dir="somewhere", **config_1)


def test_BinnedDataframe(binned_df_1, tmpdir):
    assert binned_df_1.name == "binned_df_1"
    assert len(binned_df_1._binnings) == 2
    assert binned_df_1._bin_dims[0] == "MET_px"
    # length of bin lists = nbins plus 2 for +/-inf
    assert len(binned_df_1._binnings[0]) == 29 + 2
    # length of bin lists = len(edges) - 1 plus 2 for +/-inf
    assert len(binned_df_1._binnings[1]) == 3 - 1 + 2
    assert len(binned_df_1._weights) == 1


@pytest.fixture
def binned_df_2(tmpdir, config_2):
    return bdf.BinnedDataframe("binned_df_2", out_dir="somewhere", **config_2)


@pytest.fixture
def binned_df_2_weightData(tmpdir, config_2):
    return bdf.BinnedDataframe("binned_df_2_weightData", out_dir="somewhere",
                               weight_data=True, **config_2)


def test_BinnedDataframe_run_mc(binned_df_1, tmpdir, infile):
    chunk = FakeBEEvent(infile, "mc")
    collector = binned_df_1.collector()

    binned_df_1.event(chunk)
    dataset_readers_list = (("test_dataset", (binned_df_1,)),)
    results = collector._prepare_output(dataset_readers_list)

    totals = results.sum()
    # Based on: events->Draw("Jet_Py", "", "goff")
    assert totals["n"] == 4616

    # Based on:
    # events->Draw("EventWeight * (Jet_Py/Jet_Py)>>htemp", "", "goff")
    # htemp->GetMean() * htemp->GetEntries()
    assert totals["EventWeight:sumw"] == pytest.approx(231.91339)

    coll_results = collector.collect(dataset_readers_list, writeFiles=False)

    totals = coll_results.sum()
    # Based on: events->Draw("Jet_Py", "", "goff")
    assert totals["n"] == 4616

    # Based on:
    # events->Draw("EventWeight * (Jet_Py/Jet_Py)>>htemp", "", "goff")
    # htemp->GetMean() * htemp->GetEntries()
    assert totals["EventWeight:sumw"] == pytest.approx(231.91339)


def test_BinnedDataframe_run_data(binned_df_2, tmpdir, infile):
    chunk = FakeBEEvent(infile, "data")
    binned_df_2.event(chunk)

    collector = binned_df_2.collector()
    dataset_readers_list = (("test_dataset", (binned_df_2,)),)
    results = collector._prepare_output(dataset_readers_list)

    totals = results.sum()
    # Based on: events->Draw("Jet_Py", "", "goff")
    assert totals["n"] == 4616


def test_BinnedDataframe_run_dataWeighted(binned_df_2_weightData, tmpdir, infile):
    chunk = FakeBEEvent(infile, "data")
    binned_df_2_weightData.event(chunk)

    collector = binned_df_2_weightData.collector()
    dataset_readers_list = (("test_dataset", (binned_df_2_weightData,)),)
    results = collector._prepare_output(dataset_readers_list)

    totals = results.sum()
    # Based on: events->Draw("Jet_Py", "", "goff")
    assert totals["n"] == 4616
    assert totals["weighted:sumw"] == pytest.approx(231.91339)


def test_BinnedDataframe_run_twice(binned_df_1, tmpdir, infile):
    chunk = FakeBEEvent(infile, "mc")
    collector = binned_df_1.collector()

    binned_df_1.event(chunk)
    binned_df_1.event(chunk)

    dataset_readers_list = (("test_dataset", (binned_df_1,)),)
    results = collector._prepare_output(dataset_readers_list)

    totals = results.sum()
    # Based on: events->Draw("Jet_Py", "", "goff")
    assert totals["n"] == 4616 * 2

    # Based on:
    # events->Draw("EventWeight * (Jet_Py/Jet_Py)>>htemp", "", "goff")
    # htemp->GetMean() * htemp->GetEntries()
    assert totals["EventWeight:sumw"] == pytest.approx(231.91339 * 2)


@pytest.fixture
def run_twice_data_mc(config_1, infile, observed):
    chunk_mc = FakeBEEvent(infile, "mc")
    chunk_data = FakeBEEvent(infile, "data")
    config_1["observed"] = observed

    binned_dfs = [make_binned_df_1(config_1) for _ in range(4)]
    binned_dfs[0].event(chunk_mc)
    binned_dfs[1].event(chunk_mc)
    binned_dfs[2].event(chunk_data)
    binned_dfs[3].event(chunk_data)

    return binned_dfs[0], (("test_mc", (binned_dfs[0], binned_dfs[1])),
                           ("test_data", (binned_dfs[2], binned_dfs[3])))


@pytest.mark.skipif(int(pd.__version__.split(".")[0]) < 1, reason="requires Pandas 1.0 or higher")
@pytest.mark.parametrize("dataset_col", [True, False])
@pytest.mark.parametrize("pad_missing", [True, False])
@pytest.mark.parametrize("observed", [True, False])
def test_binneddataframe_run_twice_data_mc(run_twice_data_mc, dataset_col, pad_missing, observed):
    binned_df_1, dataset_readers_list = run_twice_data_mc
    binned_df_1._pad_missing = pad_missing
    binned_df_1._dataset_col = dataset_col
    collector = binned_df_1.collector()
    results = collector._prepare_output(dataset_readers_list)

    assert results.index.nlevels == 2 + int(dataset_col)
    if pad_missing or not observed:
        length = (4 * 31)
    elif observed:
        length = 111

    length *= 1 + int(dataset_col)
    assert len(results) == length

    totals = results.sum()
    # Based on: events->Draw("Jet_Py", "", "goff")
    assert totals["n"] == 4616 * 4

    # Based on:
    # events->Draw("EventWeight * (Jet_Py/Jet_Py)>>htemp", "", "goff")
    # htemp->GetMean() * htemp->GetEntries()
    assert totals["EventWeight:sumw"] == pytest.approx(231.91339 * 2)


@pytest.fixture
def binned_df_3(tmpdir, config_3):
    return bdf.BinnedDataframe("binned_df_3", out_dir="somewhere", **config_3)


def test_BinnedDataframe_numexpr(binned_df_3, tmpdir):
    assert binned_df_3.name == "binned_df_3"
    assert len(binned_df_3._binnings) == 1
    assert binned_df_3._bin_dims[0] == "sqrt(Electron_Px**2 + Electron_Py**2)"
    assert binned_df_3._binnings[0][1].left == 0.0
    assert binned_df_3._binnings[0][-1].left == 200
    # length of bin lists = nbins plus 2 for +/-inf
    assert len(binned_df_3._binnings[0]) == 2 * 10 + 2
    assert len(binned_df_3._weights) == 1


def test_BinnedDataframe_user_var_run(config_3, tmpdir, full_wrapped_tree):
    config_4 = copy.deepcopy(config_3)
    config_4["binning"][0]["in"] = "Electron_Pt"
    binned_df_4 = bdf.BinnedDataframe("binned_df_4", out_dir="somewhere", **config_4)

    chunk = FakeBEEvent(full_wrapped_tree, "mc")
    px, py = chunk.tree.arrays(["Electron_Px", "Electron_Py"], outputtype=tuple)
    pt = np.hypot(px, py)
    chunk.tree.new_variable("Electron_Pt", pt)
    collector = binned_df_4.collector()

    binned_df_4.event(chunk)
    dataset_readers_list = (("test_dataset", (binned_df_4,)),)
    results = collector._prepare_output(dataset_readers_list)

    bin_centers = pd.IntervalIndex(results.index.get_level_values('electron_pT')).mid
    mean = np.sum((bin_centers[1:-1] * results['n'][1:-1]) / results['n'][1:-1].sum())
    assert mean == pytest.approx(44.32584)


def test_BinnedDataframe_numexpr_run_mc(binned_df_3, tmpdir, infile):
    chunk = FakeBEEvent(infile, "mc")
    collector = binned_df_3.collector()

    binned_df_3.event(chunk)
    dataset_readers_list = (("test_dataset", (binned_df_3,)),)
    results = collector._prepare_output(dataset_readers_list)

    bin_centers = pd.IntervalIndex(results.index.get_level_values('electron_pT')).mid
    mean = np.sum((bin_centers[1:-1] * results['n'][1:-1]) / results['n'][1:-1].sum())
    assert mean == pytest.approx(44.32584)


def test_BinnedDataframe_numexpr_similar_branch(binned_df_3, tmpdir, full_wrapped_tree):
    chunk = FakeBEEvent(full_wrapped_tree, "mc")
    new_var = 2 * chunk.tree.array("Electron_Px")
    chunk.tree.new_variable("tron_Px", new_var)
    collector = binned_df_3.collector()

    binned_df_3.event(chunk)
    dataset_readers_list = (("test_dataset", (binned_df_3,)),)
    results = collector._prepare_output(dataset_readers_list)

    bin_centers = pd.IntervalIndex(results.index.get_level_values('electron_pT')).mid
    mean = np.sum((bin_centers[1:-1] * results['n'][1:-1]) / results['n'][1:-1].sum())
    assert mean == pytest.approx(44.32584)


def test_explode():
    df = pd.DataFrame({'A': [[1, 2, 3], [9], [], [3, 4]], 'B': 1})
    exploded = bdf.explode(df)
    assert len(exploded) == 6
    assert all(df.B == 1)
    assert not df.isnull().any().any()

    df["C"] = df.A.copy()
    exploded = bdf.explode(df)
    assert len(exploded) == 6
    assert all(df.B == 1)
    assert all(df.A == df.C)

    df["D"] = [[1], [3], [4, 5], []]
    with pytest.raises(ValueError) as e:
        exploded = bdf.explode(df)
    assert "different jaggedness" in str(e)

    df2 = pd.DataFrame({'A': [[np.arange(i + 1) + j for i in range((j % 2) + 1)] for j in range(4)],
                        'B': np.arange(4)[::-1],
                        })
    exploded = bdf.explode(df2)
    assert len(exploded) == 8

    df = pd.DataFrame({'number': [1, 8, 3], 'string': ['one', 'eight', 'three']})
    exploded = bdf.explode(df)
    assert len(exploded) == 3

    df["list"] = [list(range(i)) for i in df.number]
    exploded = bdf.explode(df)
    assert len(exploded) == 1 + 8 + 3
    assert np.array_equal(exploded.list, [0, 0, 1, 2, 3, 4, 5, 6, 7, 0, 1, 2])

    exploded = bdf.explode(pd.DataFrame(columns=["one", "two", "3"]))
    assert exploded.empty is True

    df.index = np.arange(len(df)) + 100
    exploded = bdf.explode(df)
    assert len(exploded) == 1 + 8 + 3
    assert np.array_equal(exploded.list, [0, 0, 1, 2, 3, 4, 5, 6, 7, 0, 1, 2])


def test_densify_dataframe_integers():
    index = [("one", 1), ("one", 3), ("two", 2), ("three", 1), ("three", 2)]
    index = pd.MultiIndex.from_tuples(index, names=["foo", "bar"])
    df = pd.DataFrame({'A': np.arange(5, 0, -1), 'B': list("abcde")}, index=index)
    out_df = bdf.densify_dataframe(df, {"bar": list(range(1, 4))})

    assert len(out_df) == 9
    assert out_df.loc[("one", 2)].isna().all()
    assert out_df.loc[("two", 1)].isna().all()
    assert out_df.loc[("two", 3)].isna().all()
    assert out_df.loc[("three", 3)].isna().all()


def test_densify_dataframe_intervals():
    index = [("one", 1), ("one", 3), ("two", 2), ("three", 1), ("three", 2)]
    index = [(a, pd.Interval(b, b + 1)) for a, b in index]
    index = pd.MultiIndex.from_tuples(index, names=["foo", "bar"])
    df = pd.DataFrame({'A': np.arange(5, 0, -1), 'B': list("abcde")}, index=index)
    out_df = bdf.densify_dataframe(df, {"bar": pd.IntervalIndex.from_breaks(range(1, 5))})
    print(df)
    print(out_df)

    assert len(out_df) == 9
    assert out_df.loc[("one", pd.Interval(2, 3))].isna().all()
    assert out_df.loc[("two", pd.Interval(1, 2))].isna().all()
    assert out_df.loc[("two", pd.Interval(3, 4))].isna().all()
    assert out_df.loc[("three", pd.Interval(3, 4))].isna().all()



import pytest
import numpy as np
import fast_carpenter.define.systematics as fast_syst
from ..conftest import FakeBEEvent, Namespace


@pytest.fixture
def weights_1():
    return dict(no_variation={"nominal": "NoVariation"},
                varied={"nominal": "Varied1", "up": "Varied1Up", "down": "Varied1__DOWN"})


@pytest.fixture
def systematic_variations_1(weights_1):
    return fast_syst.SystematicWeights("systematic_variations_1", "somewhere", weights_1)


class FakeTree(Namespace):
    def array(self, attr, **kwargs):
        return getattr(self, attr)

    def new_variable(self, name, variable):
        setattr(self, name, variable)


@pytest.fixture
def fake_file():
    tree = FakeTree(Varied1=np.arange(3),
                    Varied1Up=np.arange(3) * 1.5,
                    Varied1__DOWN=np.arange(3) * 0.75,
                    NoVariation=np.full(3, 4),
                    )

    return tree


def test_systematic_variations_1_mc(systematic_variations_1, fake_file):
    chunk = FakeBEEvent(fake_file, "mc")
    systematic_variations_1.event(chunk)

    assert all(chunk.tree.weight_nominal == [0, 4, 8])
    assert all(chunk.tree.weight_varied_up == [0, 6., 12.])
    assert all(chunk.tree.weight_varied_down == [0, 3., 6.])


def test_normalize_one_variation():
    cfg = dict(nominal="one", up="two")
    stage_name = "test_normalize_one_variation"
    out = fast_syst._normalize_one_variation(stage_name, cfg, "test1", tuple())
    assert len(out) == 2
    assert out["nominal"] == "one"
    assert out["up"] == "two"

    with pytest.raises(fast_syst.BadSystematicWeightsConfig) as e:
        fast_syst._normalize_one_variation(stage_name, dict(a="bad"), "test2", tuple())
    assert "nominal" in str(e)

    with pytest.raises(fast_syst.BadSystematicWeightsConfig) as e:
        fast_syst._normalize_one_variation(stage_name, dict(nominal="one", a="bad"), "test3", tuple())
    assert "unknown key" in str(e)

    out = fast_syst._normalize_one_variation(stage_name, dict(nominal="one", a="bad"), "test3", tuple("a"))
    assert len(out) == 2
    assert out["nominal"] == "one"
    assert out["a"] == "bad"

    out = fast_syst._normalize_one_variation(stage_name, "just_a_string", "test4", tuple())
    assert len(out) == 1
    assert out["nominal"] == "just_a_string"


def test_build_variations():
    pileup = dict(nominal="PILEUP", up="PILEUP_UP", down="PILEUP_DOWN")
    isolation = dict(nominal="Iso", up="IsoUp")
    another = dict(nominal="Blahblah", left="BlahblahLeft")

    all_vars = dict(pileup=pileup, isolation=isolation, another=another)
    stage_name = "test_build_variations"
    formulae = fast_syst._build_variations(stage_name, all_vars, out_fmt="Weight_{}_test")
    assert len(formulae) == 5
    assert isinstance(formulae, list)
    assert all(map(lambda x: isinstance(x, dict) and len(x) == 1, formulae))
    formulae = {list(i.keys())[0]: list(i.values())[0] for i in formulae}

    def prep(string):
        return set(string.split("*"))

    assert prep(formulae["Weight_nominal_test"]) == prep("(PILEUP)*(Iso)*(Blahblah)")
    assert prep(formulae["Weight_pileup_up_test"]) == prep("(PILEUP_UP)*(Iso)*(Blahblah)")
    assert prep(formulae["Weight_pileup_down_test"]) == prep("(PILEUP_DOWN)*(Iso)*(Blahblah)")
    assert prep(formulae["Weight_isolation_up_test"]) == prep("(PILEUP)*(IsoUp)*(Blahblah)")
    assert prep(formulae["Weight_another_left_test"]) == prep("(PILEUP)*(Iso)*(BlahblahLeft)")



import pytest
import numpy as np
import pandas as pd
import fast_carpenter.summary.binning_config as mgr
from . import dummy_binning_descriptions as binning


def test_create_one_region():
    cfg = {"_" + k: v for k, v in binning.bins_nmuon.items()}
    _in, _out, _bins, _index = mgr.create_one_dimension("test_create_one_region", **cfg)
    assert _in == "NMuon"
    assert _out == "nmuon"
    assert _index is None
    assert _bins is None


def test_create_one_dimension_aT():
    cfg = {"_" + k: v for k, v in binning.bins_met_px.items()}
    _in, _out, _bins, _index = mgr.create_one_dimension("test_create_one_dimension_aT", **cfg)
    assert _in == "MET_px"
    assert _out == "met_px"
    assert _index is None
    assert isinstance(_bins, pd.IntervalIndex)
    assert np.all(_bins.left[1:] == np.linspace(0, 100, 30))
    assert _bins[0].left == float("-inf")
    assert _bins[-1].right == float("inf")


def test_create_one_dimension_HT():
    cfg = {"_" + k: v for k, v in binning.bins_py.items()}
    _in, _out, _bins, _index = mgr.create_one_dimension("test_create_one_dimension_HT", **cfg)
    assert _in == "Jet_Py"
    assert _out == "py_leadJet"
    assert _index == 0
    assert isinstance(_bins, pd.IntervalIndex)
    assert np.all(_bins.left[1:] == [0, 20, 100])
    assert _bins[0].left == float("-inf")
    assert _bins[-1].right == float("inf")


def test_create_binning_list():
    ins, outs, bins = mgr.create_binning_list("test_create_binning_list", [binning.bins_nmuon, binning.bins_met_px])
    assert ins == ["NMuon", "MET_px"]
    assert outs == ["nmuon", "met_px"]
    assert len(bins) == 2
    assert bins[0] is None


def test_create_weights_list():
    name = "test_create_weights_list"
    weights = mgr.create_weights(name, binning.weight_list)
    assert len(weights) == 1
    assert weights["EventWeight"] == "EventWeight"


def test_create_weights_dict():
    name = "test_create_weights_dict"
    weights = mgr.create_weights(name, binning.weight_dict)
    assert len(weights) == 1
    assert weights["weighted"] == "EventWeight"


def test_create_file_format_none():
    name = "test_create_file_format_none"
    file_format = mgr.create_file_format(name, None)
    assert len(file_format) == 1
    assert file_format[0]["extension"] == ".csv"
    assert file_format[0]["float_format"] == "%.17g"


def test_create_file_format_list():
    name = "test_create_file_format_list"
    file_format = mgr.create_file_format(name, binning.file_format_list)
    assert len(file_format) == 2
    assert file_format[0]["extension"] == ".pkl.compress"
    assert file_format[0]["compression"] == "gzip"
    assert file_format[1]["extension"] == ".csv"


def test_create_file_format_dict():
    name = "test_create_file_format_dict"
    file_format = mgr.create_file_format(name, binning.file_format_dict)
    assert len(file_format) == 1
    assert file_format[0]["extension"] == ".h5"
    assert file_format[0]["key"] == "df"


def test_create_file_format_scalar():
    name = "test_create_file_format_scalar"
    file_format = mgr.create_file_format(name, binning.file_format_scalar)
    assert len(file_format) == 1
    assert file_format[0]["extension"] == ".csv"


def test_duplicate_out_error():
    name = "test_duplicate_out_error"
    with pytest.raises(mgr.BadBinnedDataframeConfig) as e:
        mgr.create_binning_list(name, [binning.bins_nmuon, binning.bins_nmuon])
    assert "repeat" in str(e) and name in str(e)



import pytest
import numpy as np
from awkward0 import JaggedArray
import fast_carpenter.define.reductions as reductions


@pytest.fixture
def jagged_1():
    return JaggedArray.fromiter([[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8], [9.9, 10.0, 11.0], []])


def test_jagged_nth(jagged_1):
    get_first_second = reductions.JaggedNth(1, np.nan)
    reduced = get_first_second(jagged_1)
    assert reduced[0] == 1.1
    assert reduced[1] == 4.4
    assert np.isnan(reduced[2])
    assert reduced[3] == 7.7
    assert reduced[4] == 10.0
    assert np.isnan(reduced[5])


def test_jagged_nth_3D(jagged_1):
    fake_3d = [[np.arange(i + 1) + j
                for i in range(j % 3)]
               for j in range(5)]
    fake_3d = JaggedArray.fromiter(fake_3d)
    get_second = reductions.JaggedNth(1, np.nan)
    reduced = get_second(fake_3d)
    assert len(reduced[0]) == 0
    assert len(reduced[1]) == 1
    assert np.isnan(reduced[1])
    assert len(reduced[2]) == 2
    assert np.isnan(reduced[2][0])
    assert reduced[2][1] == 3
    assert len(reduced[3]) == 0
    assert len(reduced[4]) == 1
    assert np.isnan(reduced[4])

    get_first = reductions.JaggedNth(0, np.nan)
    reduced = get_first(fake_3d)
    assert len(reduced[0]) == 0
    assert len(reduced[1]) == 1
    assert reduced[1][0] == 1
    assert len(reduced[2]) == 2
    assert reduced[2][0] == 2
    assert reduced[2][1] == 2
    assert len(reduced[3]) == 0
    assert len(reduced[4]) == 1
    assert reduced[4] == 4


def test_jagged_nth_negative(jagged_1):
    get_first_second = reductions.JaggedNth(-1, np.nan)
    reduced = get_first_second(jagged_1)
    assert reduced[0] == 2.2
    assert reduced[1] == 4.4
    assert reduced[2] == 5.5
    assert reduced[3] == 8.8
    assert reduced[4] == 11.0
    assert np.isnan(reduced[5])


def test_jagged_nth_default_int(jagged_1):
    get_first_second = reductions.JaggedNth(1, 0, force_float=False)
    reduced = get_first_second(jagged_1)
    assert reduced[0] == 1
    assert reduced[1] == 4
    assert reduced[2] == 0
    assert reduced[3] == 7
    assert reduced[4] == 10
    assert reduced[5] == 0


def test_jagged_nth_default_float(jagged_1):
    get_first_second = reductions.JaggedNth(1, 0)
    reduced = get_first_second(jagged_1)
    assert reduced[0] == 1.1
    assert reduced[1] == 4.4
    assert reduced[2] == 0.
    assert reduced[3] == 7.7
    assert reduced[4] == 10.0
    assert reduced[5] == 0.0


def test_jagged_counts(jagged_1):
    get_counts = reductions.JaggedProperty("counts")
    reduced = get_counts(jagged_1)
    assert reduced[0] == 3
    assert reduced[1] == 2
    assert reduced[2] == 1
    assert reduced[3] == 3
    assert reduced[4] == 3
    assert reduced[5] == 0


def test_jagged_sum(jagged_1):
    get_sum = reductions.JaggedMethod("sum")
    reduced = get_sum(jagged_1)
    assert reduced[0] == pytest.approx(3.3)
    assert reduced[1] == pytest.approx(7.7)
    assert reduced[2] == pytest.approx(5.5)
    assert reduced[3] == pytest.approx(23.1)
    assert reduced[4] == pytest.approx(30.9)
    assert reduced[5] == pytest.approx(0)


def test_get_awkward_reduction():
    fourth_element = reductions.get_awkward_reduction("test_get_awkward_reduction", 3, fill_missing=np.nan)
    assert isinstance(fourth_element, reductions.JaggedNth)
    assert fourth_element.index == 3

    sum_elements = reductions.get_awkward_reduction("test_get_awkward_reduction", "sum")
    assert isinstance(sum_elements, reductions.JaggedMethod)
    assert sum_elements.method_name == "sum"

    count_elements = reductions.get_awkward_reduction("test_get_awkward_reduction", "counts")
    assert isinstance(count_elements, reductions.JaggedProperty)
    assert count_elements.prop_name == "counts"

    with pytest.raises(reductions.BadReductionConfig) as e_info:
        reductions.get_awkward_reduction("test_get_awkward_reduction", {"some": "dict"})
    assert "not a string or an int" in str(e_info.value)

    with pytest.raises(reductions.BadReductionConfig) as e_info:
        reductions.get_awkward_reduction("test_get_awkward_reduction", "non_existent_method")
    assert "Unknown method" in str(e_info.value)



bins_met_px = {"in": "MET_px", "out": "met_px", "bins": dict(nbins=29, low=0, high=100)}


bins_py = {"in": "Jet_Py", "out": "py_leadJet", "bins": dict(edges=[0, 20., 100.], overflow=True), "index": 0}

bins_electron_pT = {"in": "sqrt(Electron_Px**2 + Electron_Py**2)", "out": "electron_pT",
                    "bins": dict(nbins='20', low=0.0, high='2.0*10**2')}

bins_nmuon = {"in": "NMuon", "out": "nmuon"}


weight_list = ["EventWeight"]


weight_dict = dict(weighted="EventWeight")


file_format_list = [dict(extension=".pkl.compress", compression="gzip"), ".csv"]


file_format_dict = dict(extension=".h5", key="df")


file_format_scalar = ".csv"



from datetime import datetime
import os
import io
import pwd
import platform
import importlib
import tarfile
import yaml
import pandas as pd


_dependencies = ["pandas",
                 "ROOT",
                 "fast_flow",
                 "fast_curator",
                 "fast_plotter",
                 "fast_datacard",
                 "alphatwirl",
                 "yaml",
                 "coffea",
                 "alphatwirl",
                 "atuproot",
                 "atsge",
                 "numpy",
                 "matplotlib",
                 "pip",
                 'atpbar',
                 'mantichora',
                 'awkward0',
                 'numba',
                 'numexpr',
                 'uproot3',
                 ]


def _to_yaml(contents):
    # https://stackoverflow.com/questions/25108581/python-yaml-dump-bad-indentation
    class MyDumper(yaml.Dumper):
        def increase_indent(self, flow=False, indentless=False):
            return super(MyDumper, self).increase_indent(flow, False)

        # disable aliases and anchors, see https://github.com/yaml/pyyaml/issues/103
        def ignore_aliases(self, data):
            return True

    def replace_types(contents):
        if isinstance(contents, tuple):
            return list(contents)
        if isinstance(contents, list):
            return [replace_types(c) for c in contents]
        if isinstance(contents, dict):
            return type(contents)((k, replace_types(v)) for k, v in contents.items())
        if hasattr(contents, "__dict__"):
            return vars(contents)
        return contents

    return yaml.dump(replace_types(contents), Dumper=MyDumper, default_flow_style=False)


def _add_textfile(filename, tarball, contents):
    if not isinstance(contents, str):
        contents = _to_yaml(contents)
    data = contents.encode('utf8')
    info = tarfile.TarInfo(name=filename)
    info.size = len(data)
    info.mtime = datetime.now().timestamp()
    info.mode = 0o444
    tarball.addfile(info, io.BytesIO(data))


def write_booking(outfilename, sequence, datasets, extra_dependencies=[], **other):

    with tarfile.open(outfilename, "w:gz") as outfile:
        _add_textfile("sequence.yml", outfile, sequence)
        _add_textfile("datasets.yml", outfile, datasets)
        meta = prepare_metadata(other, extra_dependencies)
        _add_textfile("metadata.yml", outfile, meta)


def get_version(name):
    try:
        module = importlib.import_module(name)
    except ImportError:
        return "<not installed>"

    version = getattr(module, "__version__", "<version unknown>")
    return version


def get_platform_details():
    attrs = ["machine", "node", "processor", "release", "uname", "system", "architecture"]
    attrs += ["python_" + a for a in ["build", "compiler", "version", "implementation"]]
    return {a: getattr(platform, a)() for a in attrs}


def get_user_details():
    # Only reliable for *nix:
    details = pwd.getpwuid(os.getuid())
    return dict(name=details.pw_name)


def prepare_metadata(other, extra_dependencies=[]):
    metadata = other.copy()
    metadata["sys_info"] = pd.util._print_versions._get_sys_info()
    deps = _dependencies + extra_dependencies
    metadata["versions"] = {pkg: get_version(pkg) for pkg in deps}
    metadata["workdir"] = os.getcwd()
    metadata["platform"] = get_platform_details()
    metadata["execution_time_iso"] = datetime.now().isoformat()
    metadata["user"] = get_user_details()
    return metadata



import numpy as np
import re
import numexpr
import tokenize
import awkward0
import logging
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

logger = logging.getLogger(__name__)


__all__ = ["get_branches", "evaluate"]


constants = {"nan": np.nan,
             "inf": np.inf,
             "pi": np.pi,
             "e": np.e,
             }


def get_branches(cut, valid):
    valid = [v.decode("utf-8") for v in valid]

    branches = []
    string = StringIO(cut).readline
    tokens = tokenize.generate_tokens(string)
    current_branch = []
    for toknum, tokval, _, _, _ in tokens:
        if toknum == tokenize.NAME:
            if ".".join(current_branch + [tokval]) in valid:
                current_branch.append(tokval)
                continue
        if tokval == ".":
            continue
        if current_branch:
            branches.append(".".join(current_branch))
            current_branch = []
    return branches


def deconstruct_jaggedness(array, counts):
    if not isinstance(array, awkward0.array.base.AwkwardArrayWithContent):
        return array, counts

    array = array.compact()
    counts.insert(0, array.counts)
    return deconstruct_jaggedness(array.content, counts)


def reconstruct_jaggedness(array, counts):
    for count in counts:
        array = awkward0.JaggedArray.fromcounts(count, array)
    return array


class TreeToDictAdaptor():
    """
    Make an uproot tree look like a dict for numexpr
    """
    def __init__(self, tree, alias_dict, needed_variables):
        self.tree = tree
        self.aliases = alias_dict
        self.vars, self.counts = self.broadcast_variables(needed_variables)

    def broadcast_variables(self, variables):
        arrays = {}
        most_jagged = (-1, None)
        for var in variables:
            if var in constants:
                continue
            array = self.get_raw(var)
            contents, counts = deconstruct_jaggedness(array, counts=[])
            arrays[var] = (contents, counts, array)
            if len(counts) > most_jagged[0]:
                most_jagged = (len(counts), var)
        most_jagged = most_jagged[1]

        broadcast_to = arrays[most_jagged][1]
        broadcast_vars = {most_jagged: arrays[most_jagged]}
        for var, (contents, counts, raw) in arrays.items():
            if var == most_jagged:
                continue

            # Check broadcastable
            for left, right in zip(broadcast_to, counts):
                if not np.array_equal(left, right):
                    raise ValueError("Unable to broadcast all values")
            for copies in broadcast_to[len(counts):]:
                contents = np.repeat(contents, copies)

            broadcast_vars[var] = (contents, broadcast_to, raw)
        return broadcast_vars, broadcast_to

    def __getitem__(self, item):
        if item in constants:
            return constants[item]
        result = self.vars[item][0]
        return result

    def get_raw(self, item):
        if item in constants:
            return constants[item]
        full_item = self.aliases.get(item, item)
        array = self.tree.array(full_item)
        return array

    def __contains__(self, item):
        return item in self.vars

    def __iter__(self):
        for i in self.vars:
            yield i

    def apply_jaggedness(self, array):
        if self.counts is None:
            return array
        result = reconstruct_jaggedness(array, self.counts)
        return result


attribute_re = re.compile(r"([a-zA-Z]\w*)\s*(\.\s*(\w+))+")


def preprocess_expression(expression):
    alias_dict = {}
    replace_dict = {}
    for match in attribute_re.finditer(expression):
        original = match.group(0)
        alias = original.replace('.', '__DOT__')
        alias_dict[alias] = original
        replace_dict[original] = alias
    clean_expr = attribute_re.sub(lambda x: replace_dict[x.group(0)], expression)
    return clean_expr, alias_dict


def evaluate(tree, expression):
    cleaned_expression, alias_dict = preprocess_expression(expression)
    context = numexpr.necompiler.getContext({}, frame_depth=1)
    variables = numexpr.necompiler.getExprNames(cleaned_expression, context)[0]
    try:
        adaptor = TreeToDictAdaptor(tree, alias_dict, variables)
    except ValueError:
        msg = "Cannot broadcast all variables in expression: %s" % expression
        logger.error(msg)
        raise ValueError(msg)
    result = numexpr.evaluate(cleaned_expression, local_dict=adaptor)
    result = adaptor.apply_jaggedness(result)
    return result



import pytest
import numpy as np
from awkward0 import JaggedArray
from fast_carpenter import expressions


def test_get_branches(infile):
    valid = infile.allkeys()

    cut = "NMuon > 1"
    branches = expressions.get_branches(cut, valid)
    assert branches == ["NMuon"]

    cut = "NMuon_not_found > 1 and NElectron > 3"
    branches = expressions.get_branches(cut, valid)
    assert branches == ["NElectron"]


def test_evaluate(wrapped_tree):
    Muon_py, Muon_pz = wrapped_tree.arrays(["Muon_Py", "Muon_Pz"], outputtype=tuple)
    mu_pt = expressions.evaluate(wrapped_tree, "sqrt(Muon_Px**2 + Muon_Py**2)")
    assert len(mu_pt) == 100
    assert all(mu_pt.counts == Muon_py.counts)


def test_evaulate_matches_array(wrapped_tree):
    mu_px_array = wrapped_tree.array("Muon_Px") < 0.3
    mu_px_evalu = expressions.evaluate(wrapped_tree, "Muon_Px < 0.3")
    assert (mu_px_evalu == mu_px_array).all().all()


def test_evaluate_bool(full_wrapped_tree):
    all_true = expressions.evaluate(full_wrapped_tree, "Muon_Px == Muon_Px")
    assert all(all_true.all())

    mu_cut = expressions.evaluate(full_wrapped_tree, "NMuon > 1")
    ele_cut = expressions.evaluate(full_wrapped_tree, "NElectron > 1")
    jet_cut = expressions.evaluate(full_wrapped_tree, "NJet > 1")
    mu_px = expressions.evaluate(full_wrapped_tree, "Muon_Px > 0.3")
    mu_px = mu_px.pad(2)[:, 1]
    combined = mu_cut & (ele_cut | jet_cut) & mu_px
    assert np.count_nonzero(combined) == 2


def test_evaluate_dot(wrapped_tree):
    wrapped_tree.new_variable("Muon.Px", wrapped_tree.array("Muon_Px"))
    all_true = expressions.evaluate(wrapped_tree, "Muon.Px == Muon_Px")
    assert all(all_true.all())


def test_constants(infile):
    nan_1_or_fewer_mu = expressions.evaluate(infile, "where(NMuon > 1, NMuon, nan)")
    assert np.count_nonzero(~np.isnan(nan_1_or_fewer_mu)) == 289

    ninf_1_or_fewer_mu = expressions.evaluate(infile, "where(NMuon > 1, NMuon, -inf)")
    assert np.count_nonzero(np.isfinite(ninf_1_or_fewer_mu)) == 289


def test_3D_jagged(wrapped_tree):
    fake_3d = [[np.arange(i + 1) + j
                for i in range(j % 3)]
               for j in range(len(wrapped_tree))]
    fake_3d = JaggedArray.fromiter(fake_3d)
    wrapped_tree.new_variable("Fake3D", fake_3d)
    assert isinstance(fake_3d.count(), JaggedArray)
    assert all((fake_3d.copy().count() == fake_3d.count()).all())

    aliased = expressions.evaluate(wrapped_tree, "Fake3D")
    assert (aliased == fake_3d).all().all().all()

    doubled = expressions.evaluate(wrapped_tree, "Fake3D * 2")
    assert (doubled == fake_3d * 2).all().all().all()
    assert len(doubled[0, :, :]) == 0
    assert doubled[1, 0, :] == [2]
    assert doubled[2, 0, :] == [4]
    assert all(doubled[2, 1, :] == [4, 6])

    doubled = expressions.evaluate(wrapped_tree, "Fake3D + Fake3D")
    assert (doubled == fake_3d * 2).all().all().all()
    assert len(doubled[0, :, :]) == 0
    assert doubled[1, 0, :] == [2]
    assert doubled[2, 0, :] == [4]
    assert all(doubled[2, 1, :] == [4, 6])

    fake_3d_2 = [[np.arange(i + 3) + j
                  for i in range(j % 2)]
                 for j in range(len(wrapped_tree))]
    fake_3d_2 = JaggedArray.fromiter(fake_3d_2)
    wrapped_tree.new_variable("SecondFake3D", fake_3d_2)

    with pytest.raises(ValueError) as e:
        expressions.evaluate(wrapped_tree, "SecondFake3D + Fake3D")
    assert "Cannot broadcast" in str(e)


@pytest.mark.parametrize('input, expected', [
    ("Muon.Px > 30", ("Muon__DOT__Px > 30", {'Muon__DOT__Px': 'Muon.Px'})),
    ("events.Muon.Px > 30", ("events__DOT__Muon__DOT__Px > 30",
                             {'events__DOT__Muon__DOT__Px': 'events.Muon.Px'})),
    ('l1CaloTowerTree.L1CaloTowerTree.L1CaloTower.et > 50',
        ('l1CaloTowerTree__DOT__L1CaloTowerTree__DOT__L1CaloTower__DOT__et > 50',
         {'l1CaloTowerTree__DOT__L1CaloTowerTree__DOT__L1CaloTower__DOT__et':
             'l1CaloTowerTree.L1CaloTowerTree.L1CaloTower.et'}))
])
def test_preprocess_expression(input, expected):
    # note: maybe hypothesis.strategies.from_regex is better than parametrize
    clean_expr, alias_dict = expressions.preprocess_expression(input)
    assert clean_expr == expected[0]
    assert alias_dict == expected[1]


def test_broadcast(wrapped_tree):
    expressions.evaluate(wrapped_tree, "NJet * Jet_Py + NElectron * Jet_Px")

    with pytest.raises(ValueError):
        expressions.evaluate(wrapped_tree, "Jet_Py + Muon_Px")



import pytest
import fast_carpenter.define.variables as fast_vars


@pytest.fixture
def variable_simple():
    define = "sqrt(MET_px**2 + MET_py**2)"
    return define


@pytest.fixture
def variable_jagged():
    define = dict(formula="sqrt(Electron_Px**2 + Electron_Py**2)")
    return define


@pytest.fixture
def variable_jagged_reduce():
    define = dict(formula="sqrt(Electron_Px**2 + Electron_Py**2)", reduce="max")
    return define


@pytest.fixture
def variable_jagged_mask():
    define = dict(formula="sqrt(Electron_Px**2 + Electron_Py**2)",
                  mask="Electron_Pz>0")
    return define


@pytest.fixture
def variable_jagged_mask_reduce():
    define = dict(formula="sqrt(Electron_Px**2 + Electron_Py**2)",
                  mask="Electron_Pz>0",
                  reduce="max")
    return define


def build(name, config):
    builder = fast_vars._build_one_calc("test_" + name, name, config, approach="jagged")
    builder = builder._asdict()
    builder.pop("name")
    return builder


def test_calculation(variable_simple, variable_jagged, variable_jagged_reduce,
                     variable_jagged_mask, variable_jagged_mask_reduce, infile):
    simple = fast_vars.full_evaluate(infile, **build("simple", variable_simple))
    jagged = fast_vars.full_evaluate(infile, **build("jagged", variable_jagged))
    jagged_reduce = fast_vars.full_evaluate(infile, **build("jagged_reduce", variable_jagged_reduce))
    jagged_mask = fast_vars.full_evaluate(infile, **build("jagged_mask", variable_jagged_mask))
    jagged_mask_reduce = fast_vars.full_evaluate(infile, **build("jagged_mask_reduce", variable_jagged_mask_reduce))

    assert len(simple) == len(infile)
    assert len(jagged) == len(infile)
    assert len(jagged_reduce) == len(infile)
    assert len(jagged_mask) == len(infile)
    assert len(jagged_mask_reduce) == len(infile)
    assert all(jagged_mask_reduce <= jagged_reduce)
